;;; objc eggdoc

(use eggdoc eggdoc-zb)

;;; Document

(define doc
  `((eggdoc:begin
     (name "objc")
     (description "Scheme to Objective C bridge.")
     (author (div (a (@ (href ,zbigniew-homepage))
                     "Zbigniew")
                  ", "
                  (a (@ href "mailto:felix@call-with-current-continuation.org")
                     "Felix")
                  "."))
                 
     (history (version "0.4" (span "Class proxies, ivar refcounting, ivar Scheme objects, objc:wrap/unwrap, (@ ...) syntax, hyphenated selectors, callback optimization. "
                                   (url "http://3e8.org/zb/eggs/objc-04-changes.txt"
                                        "Detailed changelog.")))
              (version "0.3" "Enable struct support; change to ns: prefix.")
              (version "0.2" "Add objc:class-of and rename extension")
              (version "0.1" "Initial release"))
     (usage)
     (download "objc.egg")

     (documentation
      (p "This egg provides a basic interface to Objective C from Scheme.  You can invoke class and instance methods, access instance variables, and define Objective C classes directly in Scheme.")

;;; Method invocation      
      (subsection "Method invocation"
       (group
        (macro "(objc:send RECEIVER KEYWORD1 ARGUMENT1 ...)"
          (p "Sends the message KEYWORD1 ARGUMENT1 ... to RECEIVER, which is an objc:instance or objc:class object.  This follows the normal Objective C syntax, so the call " (tt "(objc:send myRect setWidth: 15.0 height: 20.0)") " invokes the method " (tt "setWidth:height:") " on myRect, with arguments 15.0 and 20.0.  If the method has no arguments, use a symbol instead of a keyword: " (tt "(objc:send NSScanner alloc)") "."))
        (macro "(objc:send/safe RECEIVER KEYWORD1 ARGUMENT1 ...)"
          (p "Identical to " (tt "(objc:send RECEIVER KEYWORD1 ARGUMENT1 ...)") ", but allows callbacks from Objective C into Scheme.  Safe calls are required when invoking any Objective C method that is either implemented in Scheme, or could itself invoke a method defined in Scheme."))
        (definition (signatures (signature "macro" "(objc:send/maybe-safe RECEIVER KEYWORD1 ARGUMENT1 ...)")
                                (signature "parameter" "(objc:optimize-callbacks) [default: #t]"))
          (p "Identical to " (tt "objc:send/safe") " when the object or class (or any superclass) is implemented in Scheme, and to " (tt "objc:send") " otherwise.  In general, this greatly improves invocation time for pure Objective C classes and objects, has a negligible impact on Scheme classes, and incurs a 10% penalty for Scheme instances.  This optimization can be disabled by setting the parameter above to " (tt "#f") (& "mdash") "in which case a safe call will always be used."))
        (macro "(@ RECEIVER KEYWORD ARGUMENT ...)"
          (p "An abbreviation for " (tt "(objc:send/maybe-safe RECEIVER KEYWORD ARGUMENT ...)") ".  The author prefers this form over the older " (tt "@[...]") " form, described next."))
        (definition (signature "read-syntax" "@[RECEIVER KEYWORD ARGUMENT ...]")
          (p "Same as " (tt "objc:send/maybe-safe") ".  If the receiver is prefixed with the " (tt "unsafe:") " keyword, then this form expands into a " (tt "(objc:send ...)") " expression instead.  Prefixing with " (tt "safe:") " guarantees an " (tt "objc:send/safe") " call.")))
       (p "For enhanced readability, the bridge accepts hyphenated selector keywords and translates them into their Objective C counterparts by uppercasing any character after a hyphen, then removing the hyphens.  For example, the following are equivalent:")
       (pre " (@ NSDictionary dictionary-with-contents-of-file: name)\n"
            " (@ NSDictionary dictionaryWithContentsOfFile: name)"))

     
;;; Instances      
      (subsection "Instances"
       (group
             (record "objc:instance"
                     (p "A wrapper for an instance of an Objective C class.  The object's " (tt "description") " method determines how this record is displayed."))
             (procedure "(objc:class-of ID)"
               (p "Return the class of objc:instance ID, obtained by sending ID a class message.  This procedure also accepts a class object, but the result will generally be the same object."))
             (procedure "(objc:instance->pointer OBJ)"
                        (p "Return the raw pointer associated with objc:instance OBJ."))
             (procedure "(objc:pointer->instance ptr)"
                        (p "Create an objc:instance from a raw instance pointer (an " (tt "id") ").  Implicitly retains the object, releasing it when the objc:instance is finalized."))))

;;; Strings
      (subsection "Strings"
       (group
        (procedure "(objc:nsstring STRING)"
                   (p "Constructs a new NSString from STRING.  Currently assumes UTF8 encoding."))
        (definition (signature "read-syntax" "@\"...\"")
          (p "Equivalent to " (tt "(objc:nsstring \"...\")") "."))
        (procedure "(objc:nsstring->string STRING)"
                   (p "Converts an NSString into a Scheme string."))))

;;; Instance variables
      (subsection "Instance variables"
        (group
         (procedure "(objc:ivar-ref OBJECT NAME)"
           (p "Returns the value of OBJECT's instance variable NAME (which should be a string).")
           (p (a (@ (href "#type-conversions")) "Type conversion")
              " is performed on the result, based on the ivar's type.  When the ivar refers to a Scheme object (i.e., is a " (url "#type-qualifiers" (tt "#:slot") " or " (tt "#:wrapper") ")") ", the object is returned transparently."))
         (definition
           (signatures (signature "procedure" "(objc:ivar-set! OBJECT NAME VALUE)")
                       (signature "procedure" "(set! (objc:ivar-ref OBJECT NAME) VALUE)"))
           (p "Sets the value of OBJECT's instance variable NAME to VALUE.")
           (p "Type conversion is performed on VALUE, based on the ivar type.  When the ivar is a "
              (tt "#:slot") " or a " (tt "#:wrapper") ", "
              " VALUE can be any Scheme object.")
           (p (url "#memory-management" "Reference counts are automatically managed") " for "
              (tt "id") " types."))
         (macro "(ivar-ref OBJECT NAME)"
           (p "Shorthand for " (tt "(objc:ivar-ref OBJECT (symbol->string NAME))") "."))
         (macro "(ivar-set! OBJECT NAME VALUE)"
           (p "Shorthand for " (tt "(objc:ivar-set! OBJECT (symbol->string NAME) VALUE)") "."))

         (definition (signature "read-syntax" "@VAR")
           (p "Intended for use within methods, this expands to " (tt "(objc:ivar-ref self \"VAR\")") ".")
           (p "Use " (tt "(set! @var value)") " to set an instance variable."))

         (pre "Examples: \n"
              "(objc:ivar-set! p \"x\" 3) \n"
              "(objc:ivar-ref p \"x\")     ; => 3 \n"
              "(ivar-ref p x)            ; => 3   \n"
              "(set! @x (vector 1 2 3))  ; when @x is a slot: \n"
              "@x                        ; => #(1 2 3)")))


;;; Classes
      (subsection "Classes"
       (macro "(define-objc-classes NAME ...)"
         (p "Locates the Objective C classes NAME ... and defines variables holding pointers to the class objects. NAME may be a symbol or a list of the form " (tt "(VARIABLE CLASSNAME)") ". For example, " (tt "(define-objc-classes NSTextView NSTask)") " is equivalent to")
         (pre
"(begin
  (define NSTextView (objc:string->class \"NSTextView\"))
  (define NSTask (objc:string->class \"NSTask\")))"))

       (group                  
        (record "objc:class"
                (p "A record representing an Objective C class.")
                (symbol-table
                 (describe "name" "Class name as a string.")
                 (describe "method-list" (div "List of instance methods.  The format is " (tt "((NAME . SIGNATURE) ...)") ", but may change in the future to a list of objc:method records."))
                 (describe "class-method-list" "List of class methods.")
                 (describe "super-class" "Superclass of this class (as an objc:class).")
                 (describe "meta-class" "Metaclass of this class (as an objc:class).")
                 (describe "ivar-list" "List of all instance variables in this class (as objc:raw-ivar records).")
                 (describe "ivars" "List of ivars in this class defined in Scheme (as objc:ivar records).")
                 (describe "all-ivars" "Aggregate list of objc:ivar records from the class hierarchy."))
                (p "A note on instance variables: the bridge generates " (tt "objc:ivar") " records only for ivars defined in Scheme.  However, " (tt "objc:raw-ivar") " records are available for all instance variables.  This API (" (tt "ivar-list / ivars / all-ivars") ") is new and subject to change."))

        (procedure "(objc:class->pointer CLASS)"
                   (p "Return the raw pointer associated with objc:class CLASS."))
        (procedure "(objc:pointer->class ptr)"
                   (p "Create a class from a raw class pointer (a " (tt "Class") ")."))
        (procedure "(objc:string->class STRING)"
                   (p "Look up and return the Objective C class named STRING."))

        (definition (signatures (signature "macro" "(objc:define-method CLASS RT ARGS . BODY)")
                                (signature "macro" "(objc:define-class-method CLASS RT ARGS . BODY)"))
          (p "Define an instance or class method in CLASS.")
          (symbol-table
           (describe "CLASS" "An objc:class object representing the destination class.")
           (describe "RT"    "The return type of the method.")
           (describe "ARGS"  (div (tt "((KEYWORD TYPE VAR-NAME) ...)") " or " (tt "SYMBOL")))
           (describe "BODY"  "Body of a lambda comprising the method.  The parameters visible
                            inside the lambda are self (a class or instance object), sel (the method
                            selector), and the arguments given in ARGS."))
          (p "Each list in ARGS adds a method argument of type "
             "TYPE, visible to the method body as VAR-NAME.  As in Objective C, "
             "each argument is associated with a KEYWORD and each KEYWORD is "
             "combined into a method name.  Hyphenated keywords are accepted, just like "
             "in method invocations.  A bare SYMBOL can be used as the "
             "method name (instead of a list) if no arguments are expected.")
          (p "TYPE should be a " (url "#types" "short typename") ", a symbol such as " (tt "INT") ".")
          (p "Within a method body, you can use " (tt "(@ super ...)") " to call a class or instance method of the superclass.  Note: " (tt "super") " is a reserved keyword, not a variable.")

          (p "Example:")
          (pre "
 (objc:define-method Rect VOID ((set-width: DBL my-w)
                                (height:    DBL my-h))
   (ivar-set! self w my-w)
   (ivar-set! self h my-h))")
          
          (p "Method removal is not yet implemented, but redefining a method will override the old definition."))
      
        (macro "(define-objc-class CLASS SUPERCLASS IVARS . METHODS)"
          (p "Defines CLASS (a symbol) with superclass SUPERCLASS (a symbol), instance variables IVARS, and methods METHODS.  The new classname is imported with " (tt "define-objc-classes") ".")
          (p "SUPERCLASS is looked up for you in the runtime, so it need not be imported.")
          (p "IVARS is a list of the form " (tt "((TYPE NAME) ...)") ", where TYPE is a " (url "#type-qualifiers" "type qualifier") " and NAME is a symbol representing the new variable name.  Each instance of CLASS will have a separate copy of these variables.")
          (p "METHODS are method definitions of the form " (tt "(define-[class-]method RT ARGS . BODY)") ", which are equivalent to calling " (tt "(objc:define-[class]-method CLASS RT ARGS . BODY)") " using the current CLASS.  These methods are defined in the lexical environment of the surrounding " (tt "define-objc-class") " expression.  As a simple consequence, you can surround the class definition with a " (tt "let") " statement and create \"static\" variables for the class.")
          (p "You can also use " (b "+") " as an alias for " (tt "define-class-method") " and " (b "-") " for " (tt "define-method") ".  These correspond to Objective C method definition syntax.")
          (pre
"Example:

 (define-objc-class MyPoint NSObject ((DBL x)
                                      (DBL y) 
                                      (slot: closure))
   (define-method ID init
     (print \"MyPoint init\")
     (@ super init))
   (- DBL getX @x)
   (- DBL getY @y)
   (- ID description
      (sprintf \"<MyPoint: (~a, ~a)>\" @x @y))

   (- VOID ((move-by-x: DBL a) (y: DBL b))
     (set! @x (+ a @x))
     (ivar-set! self y (+ b (ivar-ref self y))))  ;; more wordy

   (- ID ((init-with-x: DBL a) (y: DBL b))
      (let ((p (@ self init)))
        (@ p move-by-x: a y: b)
        (set! @closure (lambda (msg)
                         (cond ((eq? msg 'initial-x) 
                                (print \"MyPoint: initial x was \" a))
                               ((eq? msg 'initial-y) 
                                (print \"MyPoint: initial y was \" b)))))
        p)))

 #;1> (define p (@ (@ MyPoint alloc) init-with-x: 3.4 y: 4.5))
 MyPoint init
 #;2> (@ p move-by-x: 2 y: 3)
 #<objc-instance <MyPoint: (5.4, 7.5)>>
 #;3> ((ivar-ref p closure) 'initial-x)
 MyPoint: initial x was 3.4"))
        

        (parameter "(objc:allow-class-redefinition) [default: #t]"
          (p "If #f, an error will occur when attempting to redefine an existing class with define-objc-class.")
          (p "If #t, redefinition is allowed and a warning will be printed."))))


;;; Types                  
      (subsection (a (@ (Name "types")) "Types")   ;; "Name" capitalized to avoid conflict with eggdoc.
        (p "Objective C is a typed language, and the bridge uses these types to decide how to pass values into and out of Objective C.  Each type is represented by a specific string and the bridge provides a variable containing each type string.  The variable names and their associated Objective C types are listed below.")
        (p (tt "objc:define-method") " uses short versions of the types below, with the " (tt "objc:") " prefix removed.  For example, use " (tt "ID") " instead of " (tt "objc:ID") "." " The full names are generally used for lower-level methods such as " (tt "objc:add-method") " and " (tt "objc:set-ivars!") ".")
        
        
        (symbol-table
          (describe "objc:ID" "id")
          (describe "objc:CLASS" "Class")        
          (describe "objc:SEL" "SEL")
          (describe "objc:INT" "int")
          (describe "objc:DBL" "double")
          (describe "objc:FLT" "float")
          (describe "objc:CHR" "char")
          (describe "objc:SHT" "short")
          (describe "objc:LNG" "long")
          (describe "objc:USHT" "unsigned short")
          (describe "objc:UINT" "unsigned int")
          (describe "objc:UCHR" "unsigned char")
          (describe "objc:ULNG" "unsigned long")
          (describe "objc:BOOL" "BOOL")
          (describe "objc:PTR" "void *")
          (describe "objc:CHARPTR" "char *")
          (describe "objc:NSRECT" "NSRect")
          (describe "objc:NSSIZE" "NSSize")
          (describe "objc:NSPOINT" "NSPoint")
          (describe "objc:NSRANGE" "NSRange")
          (describe "objc:VOID" "void"))

        (a (@ Name "type-qualifiers")
            (p "Instance variables defined in " (tt "define-objc-class") " use the type qualifiers in the following table."))

        (group
         (symbol-table
          (describe "DBL, INT, etc."
                    (div "An " (tt "objc:DBL") ", " (tt "objc:INT") ", etc. type"
                         (& "mdash") "use short typenames, as in " (tt "objc:define-method") "."))
          (describe "ID" (div "An Objective C instance (objc:ID) with automatic "
                              (a (@ (href "#memory-management")) "memory management")
                              "."))
          (describe "#:outlet" "An Interface Builder outlet (objc:ID).  Memory management is not performed.")
          (describe "#:slot" "Holds a scheme object, such as a vector or closure.")

          (describe "#:wrapper" "Holds a scheme object like #:slot, but is less efficient."))))

;;; Type conversions

      (subsection (a (@ (Name "#type-conversions")) "Type conversions")
        (p "Numeric types (such as double and int) are converted to and from Scheme numbers just as in Chicken's C FFI.")
        (p "Class types are wrapped in unique objc:class records when passed to Scheme" (& "mdash")
           "see " (a (@ (href "#class-proxies")) "Class Proxies")
           " for more details.")
        (p "An " (tt "id") ", or instance, type is almost always represented as an " (tt "objc:instance") " record, which is a thin wrapper around a pointer to the Objective C object.  There is generally no automatic conversion to or from Scheme objects, even when the object has a reasonable direct representation in Scheme.  For example, you may not pass the number \"3\" to a method expecting an " (tt "id") ", even though \"3\" could be represented as an " (tt "NSNumber") ".  Conversely, an NSNumber representing \"3\" remains an " (tt "objc:instance") " when returned to Scheme.")
        (p "There is one exception to this rule: you may pass a string to any method which expects an " (tt "id") ", and it will become an " (tt "NSString") ".  Of course, you can always perform conversions manually using Objective C methods" (& "mdash") "continuing the example above, an " (tt "(@ NSNumber number-with-int: 3)") " can indeed be passed as an " (tt "id") " argument.  The author has written convenience functions for NSNumber, NSDictionary, and NSArray which reduce the drudgery of conversion.  These should be available soon; " (a (@ (href ,zbigniew-homepage)) "check here for updates") ".")
        (p "Objective C lacks a boolean type; booleans are char types where zero is false and non-zero is true.  Since (char)0 is rare, we convert it to #f when passed to Scheme, which allows Scheme predicates to work without a special test.  We also transform #t and #f to (char)1 and (char)0 when passed to Objective C.  Other character values are passed through as-is.  Unsigned char values, on the other hand, never represent booleans and aren't transformed.")
        (p "Selectors are converted to strings when passed to Scheme, and strings converted back to selectors when passed to Objective C.  Note that selectors may be wrapped in objc:selector objects in the future.")
        (p "CHARPTR (char *) types are converted to Scheme strings, but conversion to CHARPTR is disabled.")
        (group
         (p "NSRect, NSPoint, NSSize and NSRange structures can be sent to and received from Objective C.  Each is represented by a record on the Scheme side.")
         (record "ns:rect"
                 (p (tt "ns:make-rect") " is provided as an alias for the default constructor " (tt "make-ns:rect") ".  The same is true for the other records.")
                 (symbol-table (describe "x" "X coordinate of origin (NSRect.origin.x)")
                               (describe "y" "Y coordinate of origin (NSRect.origin.y)")
                               (describe "width" "Width of rectangle (NSRect.size.width)")
                               (describe "height" "Height of rectangle (NSRect.size.height)")))
         (record "ns:size"
                 (symbol-table (describe "width" "Width")
                               (describe "height" "Height")))
         (record "ns:point"
                 (symbol-table (describe "x" "X coordinate")
                               (describe "y" "Y coordinate")))
         (record "ns:range"
                 (symbol-table (describe "location" "Start index, 0-based")
                               (describe "length" "Length of range")))
         
         (p "Other than the four exceptions above, struct, union, and array types cannot be sent to or received from Objective C.")))
       
;;; Memory Management
      (subsection (a (@ (Name "memory-management")) "Memory management")
       (p "The bridge strives to handle memory management automatically.  In general terms, this is accomplished by implicitly retaining an instance object when it is passed into Scheme, and releasing it when the objc:instance is finalized.  The bridge knows that certain selectors, such as alloc and copy, by convention donate a retain reference to you, and adjusts its retain count accordingly.  Furthermore, passing an objc:instance into Objective C retains and autoreleases that object, which is necessary to ensure that short-lived objects (such as automatically created strings) remain valid until the end of a method invocation.  (Each invocation is wrapped in an autorelease pool, hence the autorelease.)")
       (p "In general, this means you don't have to worry about releasing, autoreleasing or retaining objects.  The expression " (tt "(let ((m [@ MyPoint alloc])) (void))") ", for example, incurs no memory penalty as " (tt "m") " is garbage collected like any other Scheme object.  This principle extends to NSStrings that are created by the bridge when converted from Scheme strings.  Additionally, you may return newly allocated objects from Scheme classes without autoreleasing them.")
       (p "Instance variables which hold Objective C objects"
          (& "mdash") (b "id") " types" (& "mdash")
          "are also managed automatically.  Those defined in a Scheme class are properly retained when using " (tt "objc:ivar-set!") ", and properly released in " (tt "dealloc") ".  You should use the " (tt "#:outlet") " type qualifier for Interface Builder outlets, which will turn off automatic management as IB expects.  Ivars defined in a pure Objective C class are never retained or released automatically, as these classes expect to manage memory themselves.  If you must access such a variable directly, you must manually send it retain and release messages (editor's note: the API for such is not exposed right now).")
       (p "There are some limitations.  Overriding memory-management selectors such as alloc, release, and dealloc is not supported (although alloc does appear to work fine, caveat emptor).  Sending an autorelease message has no effect due to the current implementation of method invocation.  Finally, although the author has tried to ensure automatic memory management works as advertised, certain cases (especially involving calls from Objective C to Scheme) have not been tested and may be problematic at this point."))


      (subsection (a (@ Name "class-proxies")
                     "Class proxies")
        (p "In order to implement ivar memory management and transparent access to Scheme objects stored in ivars, the bridge needs to maintain metadata for each class defined in Scheme.  Enter the class proxy, a unique " (tt "objc:class") " record for each class.  Whenever a class pointer is passed into Scheme, this corresponding proxy is looked up and returned.  This means that Scheme can store information about classes beyond that available in the Objective C runtime.")
        (p "For pure Objective C classes, this proxy is generic, and springs into being dynamically the first time the class is referenced.  It doesn't contain any extra data; it simply notes the class is pure Objective C.")
        (p "For classes defined in Scheme, the proxy is created at class definition time, and contains (amongst other things) extended instance variable data, including type qualifiers such as slot: and outlet:.  " (tt "objc:ivar-set!") " will look up the object's class, find the appropriate ivar and notice any type qualifier.")
        (p "Instance proxies per se are not (yet) implemented.  Although objects are wrapped in objc:instance records, each is merely a non-unique wrapper around a pointer and contains no extra data.  Of course, since instance variables can hold Scheme objects, you can store as many \"proxy\" objects as you like inside any instance.  On the other hand, you have to manually access this data via ivar-ref or by sending the Objective C object a message, which can be cumbersome.  In practice, it may be useful to do this transparently, having an interchangeable proxy Scheme object and Objective C instance.  This is an active area of research.")
        ;; (p "Note that these proxies are not classes in the sense of TinyCLOS or Meroon.  It would not be difficult to map a class pointer to a \"real\" class rather than a simple record, but this raises a few issues.  It would almost certainly require that object instances be mapped to instance proxies (instances of class proxies, here).  ")
        )
      
;;; Lowlevel
      (subsection "Lowlevel"
       (group
        (procedure "(objc:register-class CLASSNAME SUPERCLASS)"
          (p "Registers CLASSNAME (a string) having superclass SUPERCLASS (an objc:class) with the Objective C runtime.  An error is raised if CLASSNAME already exists."))
      
        (procedure "(objc:set-ivars! CLASS IVARS)"
          (p "Defines in CLASS (an objc:class) the instance variables in IVARS (a list of objc:instance-var records).  The offset parameter of the instance variable records is ignored.")
          (p "Warning: all old instance variables in MyClass will be removed first. Also, we don't check for conflicts with superclass instance variables.  This should be remedied in a future release.")
          (pre "Example: \n"
               "(objc:set-ivars! MyClass (list (make-objc:raw-ivar \"jimmy\" objc:INT 0) \n"
               "                               (make-objc:raw-ivar \"cammy\" objc:DBL 0)"))

        (definition (signatures
                     (signature "procedure" "(objc:add-method CLASS METHOD TYPES PROC)")
                     (signature "procedure" "(objc:add-class-method CLASS METHOD TYPES PROC)"))
          (p "Adds a class or instance method to CLASS.")
          (symbol-table
           (describe "METHOD" (div "Method name as a string (e.g. " (tt "\"setWidth:height:\"") ")"))
           (describe "TYPES"  (div "List of encoded argument type strings (such as " (tt "objc:INT") " or \"@\")."))
           (describe "PROC"   "Scheme procedure representing the method."))
          (p "The structure of the TYPES list is " (tt "(RETURN-TYPE SELF SELECTOR METHOD-ARGS...)") ".")
          (pre
      "Transformation: \n"
      "(objc:define-method MyClass DBL ((sel1: INT i) (sel2: DBL d)) \n"
      "                                (print i) (+ i d)) \n"
      "=> \n"
      "(objc:add-method MyClass \"sel1:sel2:\" (list objc:DBL objc:ID objc:SEL objc:INT objc:DBL) \n"
      "                                      (" (& "lambda") " (self sel i d) (print i) (+ i d))) \n"))

        (definition
          (signatures (signature "procedure" "(objc:wrap X)")
                      (signature "procedure" "(objc:unwrap X)"))
          (p "Wrap or unwrap the Scheme object X inside an Objective C instance (specifically, a " (tt "Scheme_Object_Wrapper") ") so that it can be passed as an " (tt "id") " type.  Essentially, these functions allow you to tunnel a Scheme object through the Objective C bridge, when both endpoints are written in Scheme.  At the moment, the resulting object cannot be accessed meaningfully from the Objective C side.")
          (p "These functions are also used to implement the " (tt "#:wrapper") " type qualifier for instance variables."))
        
        (record "objc:raw-ivar"
          (p "A record describing an Objective C instance variable as seen from the Objective C (as opposed to Scheme) side.  Returned by " (tt "objc:class-ivar-list") ", and used by " (tt "objc:set-ivars!") ".")
          (symbol-table
           (describe "name" "Name as a string.")
           (describe "type" "Type as an encoded type string.")
           (describe "offset" "Offset within class -- for debugging only.")))
        (record "objc:ivar"
                (p "A record describing Scheme's view of an Objective C instance variable.  At the moment, fields include those of " (tt "objc:raw-ivar") " with the following addition:")
                (symbol-table
                 (describe "function" "Keyword #:slot, #:wrapper, #:outlet, or #:ivar.")))
        (procedure "(with-autorelease-pool THUNK)"
                   (p "Creates an autorelease pool that lasts for the duration of the thunk.")
                   (p "A global autorelease pool is created automatically at startup, one is wrapped around the main Application Kit event loop, one is wrapped around every call to Objective C, and memory management is generally otherwise automatic.  It is considered unlikely you will have to use this, unless you send Objective C messages directly inside a foreign-lambda."))
        (procedure "(objc:import-classes-at-toplevel!)"
                   (p "Import every class visible to the runtime, as if define-objc-classes had been run on all available classes.  Useful for debugging.  Note that some (rare) classes are not derived from NSObject, and will not respond to standard NSObject selectors, may throw exceptions, or may crash if used."))
        (procedure "(objc:get-class-list)"
                   (p "Looks up and returns a list of all available classes.  At startup, the result of this call is stored in the variable " (tt "objc:classes") "."))))

      )  ;; end documentation

;;; Cocoa
     (section "Cocoa"
      (p (tt "(require-extension cocoa)"))
      (p "It is possible to create Cocoa applications using this extension.  See " (url "http://3e8.org/zb/cocoa/creating-a-cocoa-app.html" "Creating a Cocoa Application in Chicken") " for a document which walks you through implementing Apple's famous Currency Converter application.")
      (p "Also, a working application is included in this egg.  Untar the egg, change to the " (tt "tests/") " directory, and type " (tt "make") ".  An application called Temperature Converter will be built.")
      (group
       (procedure "(ns:application-main)"
         (p "Starts the main event loop of a Cocoa-based application.  If any arguments are present on the command line, they will be passed to NSApplicationMain."))
       (procedure "(ns:beep)" (p "Plays the default system sound using NSBeep."))
       (procedure "(ns:log FORMAT-STR ARGS...)"  (p "Logs a message using NSLog.  The optional " (tt "ARGS") " are interpolated into " (tt "FORMAT-STR") " as in " (tt "printf") "."))
       (procedure "(ns:rect-fill RECT)" (p "Fills the passed RECT (an " (tt "ns:rect") ") with the current color, using NSRectFill.")
       (procedure "(cocoa:run)"
         (p "Send the run message to the global NSApplication object.  This is intended to be used during debugging, to restart the main application event loop after an error has returned you to the REPL."))))
      (p "Some " (url "http://developer.apple.com/documentation/Cocoa/Reference/ApplicationKit/ObjC_classic/Functions/AppKitFunctions.html" "global Application Kit functions") ", such as NSRectFill, take structures which are passed by value.  However, the Chicken FFI only supports passing structs by pointers.  A workaround is provided in the egg; if you need to wrap such a function, see " (tt "ns:rect-fill") " in " (tt "cocoa.scm") " for an example."))

     (license ,licenses:zbigniew-bsd))))

;;; Generate eggdoc

(eggdoc->html doc)
                                 
