;;; objc Scheme<->ObjC bridge

;(use objc-support) ;; instead, use require-at-runtime in objc.setup
;(objc:import-classes-at-toplevel!)

;;; invoker macros

;; (objc TypeTest printInt: 1.1 Double: 2.2 Float: 3.3)
;; Allows target "super" -- e.g. @[super init] is transformed
;; into @[self classname:super:init].  (The classname: is required
;; because super calls the superclass of the class defining method, not
;; the superclass of self.)

(define-macro (objc:send target arg . args)
  (%objc #f target arg args))
(define-macro (objc:send/safe target arg . args)
  (%objc #t target arg args))
(define-macro (objc:send/maybe-safe target arg . args)
  (%objc 'maybe target arg args))
(define-macro (@ . args)
  `(objc:send/maybe-safe ,@args))


;; Old forms:
(define-macro (objc . args)
  `(objc:send ,@args))            ;; doesn't work (?)
(define-macro (objc/safe . args)
  `(objc:send/safe ,@args))

(define (%objc safe? target arg args)
  (let ((super? (eq? target 'super)))
    (if (null? args)
        (if super?
            `(objc:invoker ',safe? self (string-append _supersel
                                                      ,(objcify-selector (symbol->string arg))))
            `(objc:invoker ',safe? ,target ,(objcify-selector (symbol->string arg))))
        (receive (method passargs)
            (apply objc:aggregate-args (cons arg args))
          (if super?
              `(objc:invoker ',safe? self (string-append _supersel ,(objcify-selector method))
                             ,@passargs)
              `(objc:invoker ',safe? ,target ,(objcify-selector method) ,@passargs))))))

;; For parameter names, we accept actual keywords instead of symbols ending in :.
;; Thus, depending on the current keyword-style, initWithValue:, #:initWithValue, or
;; :initWithValue will be converted to "initWithValue:". 
;; Note a single argument taking no value requires a bare symbol, not a keyword.
(eval-when [compile eval]
 (set! objc:aggregate-args (lambda args
  (letrec ((keyword (lambda (ls method-name params)
		      (if (null? ls)
			  (values method-name (reverse params))
			  (param (cdr ls)
				 (string-append
                                  method-name
                                  (let ((this-method (car ls)))
                                    (cond ((keyword? this-method)
                                           (string-append (symbol->string this-method) ":"))
                                          ((symbol? this-method)
                                           (symbol->string this-method))
                                          (error 'objc "keyword expected" this-method))))
				 params))))
	   (param   (lambda (ls method-name params)
		      (if (null? ls)
			  (error 'objc "malformed method name")
			  (keyword (cdr ls)
				   method-name
				   (cons (car ls) params))))))
    (keyword args "" '())))))

;; Convert a scheme-type selector string to Objective C syntax.  This simply entails
;; uppercasing any character after a hyphen.  This is only done during macroexpansion.
(eval-when [compile eval]
  (require 'srfi-13)           
  (define (objcify-selector sel)
    (let ((pieces (string-split sel "-")))
      (for-each (lambda (s) (string-upcase! s 0 1))
                (cdr pieces))
      (apply string-append pieces))))

;;; Instance variables

;; This macro is slightly counterintuitive, as the name is expected to be a symbol
;; (and thus fixed at read time).  On the other hand, you can use objc:ivar-ref,
;; which takes a real string.
;; Note: these are macros, and cannot comply with SRFI-17.  However, (ivar-set! ...) is
;; shorter than (set! (ivar-ref ...)) anyway, and the (set! @foo 'bar) syntax still works.
(define-macro (ivar-ref obj name)
  `(objc:ivar-ref ,obj ,(symbol->string name)))
(define-macro (ivar-set! obj name val)
  `(objc:ivar-set! ,obj ,(symbol->string name) ,val))

;;; Class definitions

;;;; define-class

#|

Example transformation:

(define-objc-class MyPoint NSObject ((FLT x) (FLT y))
  (define-method FLT getX (ivar-ref self x))
  (define-method VOID ((moveByX: FLT a) (Y: FLT b))
    (ivar-set! self x (+ a (ivar-ref self x)))
    (ivar-set! self y (+ b (ivar-ref self y)))))

=>

(begin
  (if (string-to-class "MyPoint")
      (warning (conc "(define-objc-class): class already registered: " 'MyPoint))
      (objc:register-class "MyPoint" (objc:string->class "NSObject")))
  (define-objc-classes MyPoint)
  (objc:set-ivars! MyPoint (list (make-objc:raw-ivar "x" objc:FLT 0)
                                (make-objc:raw-ivar "y" objc:FLT 0)))
  (objc:define-method MyPoint FLT getX (ivar-ref self x))
  (objc:define-method MyPoint VOID ((moveByX: FLT a) (Y: FLT b))
    (ivar-set! self x (+ a (ivar-ref self x)))
    (ivar-set! self y (+ b (ivar-ref self y)))))

|#

;; Whether we warn or error out if a class is already defined.
(define objc:allow-class-redefinition (make-parameter #t))

;; The superclass will be looked up for you; it does not need to be imported.
(define-macro (define-objc-class class super ivars . methods)
  `(begin
     ;; register class
     (if (string-to-class ,(symbol->string class))
         ((if (objc:allow-class-redefinition)
              warning error)
          ,(string-append "(define-objc-class): class already registered: "
                          (symbol->string class)))
         (objc:register-class ,(symbol->string class)
                              (objc:string->class ,(symbol->string super))))
     ;; import class
     (define-objc-classes ,class)
     ;; set instance variables
     (objc:set-ivars! ,class
                     (list ,@(map (lambda (ivar)
                                    (let ((type (car ivar))
                                          (name (cadr ivar)))
                                      `(make-objc:raw-ivar ,(symbol->string name)
                                                           ,(macro:type->encoding type)
                                                           0)))
                                  ivars)))
     ;; add methods
     ,@(map (lambda (method)
              (let ((definer (case (car method)
                               ((define-method -) 'objc:define-method)
                               ((define-class-method +) 'objc:define-class-method)
                               (else (error "invalid method definition keyword" (car method))))))
                `(,definer ,class ,@(cdr method))))
            methods)))
  
;;;; define-method

;; Transformation:
;; (objc:define-method MyClass DBL ((sel1: INT i) (sel2: DBL d))
;;                                 (print i) (+ i d))
;; =>
;; (objc:add-method MyClass "sel1:sel2:" (list objc:DBL objc:ID objc:SEL objc:INT objc:DBL)
;;                                       (lambda (self sel i d) (print i) (+ i d)))
(define-macro (objc:define-method class rt args . body)
  (%define-method #f class rt args body))
(define-macro (objc:define-class-method class rt args . body)
  (%define-method #t class rt args body))

(define (macro:type->encoding x)   ;; internal
  (cond ((symbol? x) (string->symbol
                      (string-append "objc:" (symbol->string x))))
        (else x)))

;; Discrepancy: we compute _classname at macroexpansion time from the class -symbol-, but
;; objc:add-method creates the super selector at runtime from the actual registered class name.

(eval-when [compile eval]
 (set! %define-method (lambda (class? class rt args body)  ;; internal helper function
  (define (add-method-body method-name types names)
    (let ((self-type  (if class? 'CLASS 'ID))
          (add-method (if class? 'objc:add-class-method
                          'objc:add-method)))
      `(,add-method ,class
                    ,(objcify-selector method-name)
                    (list ,@(map (cut macro:type->encoding <>) (apply list rt self-type 'SEL types)))
                    (let ((_supersel (string-append ,(symbol->string class) ":super:")))
                      ;; _supersel is a hidden variable used by @[super..]
                      (lambda (self sel ,@names) ,@body)))))

  (if (pair? args)
      (let* ((args  (apply map list args))  ;; '((sel: type name) ...) =>
             ;; '((sel: ...) (type ...) (name ...))
             (sels  (car args))
             (types (cadr args))
             (names (caddr args))
             (method-name (apply string-append
                                 (map (lambda (x) (string-append (keyword->string x) ":"))
                                      sels))))
        (add-method-body method-name types names))
      (let ((method-name (if (keyword? args)
                             (error 'objc:define-method "argument required for selector" args)
                             (symbol->string args))))
        (add-method-body method-name '() '()))))))
    
;; Note: type is normally a keyword; objc: will be prepended (e.g. objc:ID).  If
;; not a keyword, it is pasted verbatim so you can e.g. pass an encoded typestring.

;;; Importing classes

(define-macro (define-objc-classes . names)
  `(begin
     ,@(map (lambda (name)
	      (cond ((symbol? name)
		     `(define ,name (objc:string->class ,(->string name)) ))
		    ((and (list? name) (= (length name) 2) (symbol? (car name)))
		     `(define ,(car name) (objc:string->class ,(->string (cadr name)))) )
		    (else (syntax-error 'define-objc-classes "invalid class name" name))))
	    names) ) )

;;; Read syntax

;; Felix's @[] read syntax implementation, with one tweak: all calls are maybe-safe
;; unless prefixed by unsafe: or safe:.
;; @[target sel1: x sel2: y] => (objc sel1: x sel2: y)
;; @"..."                    => creates NSString from "..."
;; @foo                      => (ivar-ref self foo)
(set-read-syntax! 
 #\@
 (let ([terminating-characters '(#\, #\; #\) #\] #\{ #\} #\')])
   (lambda (p)
     (let ((c (peek-char p)))
       (if (or (char-whitespace? c) (memq c terminating-characters))
	   '@
	   (let ((x (read p)))
	     (cond ((keyword? x) (string->keyword (string-append "@" (keyword->string x))))
		   ((symbol? x) ;(string->symbol (string-append "@" (symbol->string x)
                    `(objc:ivar-ref self ,(symbol->string x)))  ;; Can't use macro due to SRFI-17
		   ((string? x) `(force (delay (objc:nsstring ,x))))
		   ((pair? x) 
		    (cond ((eq? #:safe (car x))
                           `(objc:send/safe ,@(cdr x)))
                          ((eq? #:unsafe (car x))
                           `(objc:send ,@(cdr x)))
                          (else
                           `(objc:send/maybe-safe ,@x))))
		   (else (error "invalid read syntax for `@'" c)) ) ) ) ) ) ) )

