;;; Chicken ObjC bridge -- basic proxies

(require-extension objc-base)

;;; Class definition macro

;; objc:ivar-ref and objc:ivar-set! use the instance variable type qualifiers
;; to manage memory correctly and transparently provide access to scheme objects.

;; The type qualifiers:
;;    #:slot    An ID containing a GC root pointer to a scheme object.
;;               Scheme objects are transparently read from and assigned to these.
;;    #:wrapper An ID containing a Scheme_Object_Wrapper (which contains a GC root).
;;               Same as #:slot, with higher overhead.  Uses objc:wrap and unwrap.
;;    #:outlet  An ID whose reference count is not managed.
;;               Use these for Interface Builder outlets.
;;    ID        An ID whose reference count is automatically managed.
;;               Use these for normal Objective C instances.
;;    DBL, &c.  A regular Objective C instance variable.

;; NB!! In the interpreter, if you assign a closure to a slot: or
;; wrapper: var, it will capture self.  There is then a deadlock: self
;; cannot be released until the closure is released; the closure
;; cannot be released until the GC root is deleted, the GC root can't
;; be deleted until objc-self's dealloc is called; objc-self's dealloc
;; can't be called until its retain count is 0; and retain count
;; remains at 1 until self is released.
;;      The compiler however is smart enough to know when the closure
;; does not depend on self.  Finalization works fine in that case.

;; The superclass will be looked up for you; it does not need to be imported.
(define-macro (define-objc-class class super ivars . methods)
  (let ((instance-variables (gensym)))
  `(begin
     ;; register class
     (if (string-to-class ,(symbol->string class))
         ((if (objc:allow-class-redefinition)
              warning error)
          ,(string-append "(define-objc-class): class already registered: "
                          (symbol->string class)))
         (objc:register-class ,(symbol->string class)
                              (objc:string->class ,(symbol->string super))))
     ;; import class
     (define-objc-classes ,class)
     (set! (objc:class-objc? ,class) #f)   ;; This class is not pure ObjC.

     ;; set instance variables
     (let ((,instance-variables
            (list ,@(map (lambda (ivar)
                           (let ((qualified-ID? (memq (car ivar)
                                                      '(slot: wrapper: outlet:))))
                             (let ((name        (cadr ivar))
                                   (type        (if qualified-ID? 'ID (car ivar)))
                                   (function    (if qualified-ID? (car ivar) ivar:)))
                               `(make-objc:ivar ,(symbol->string name)
                                                ,(macro:type->encoding type)
                                                0
                                                ,function))))
                         ivars))))
       ;; Set instance vars on the Objective C side...
       (objc:set-ivars! ,class (map objc:ivar->raw ,instance-variables))
       ;; ... and in the Scheme class proxy.
       (set! (objc:class-ivars ,class)
             (map (lambda (x)
                    (cons (objc:ivar-name x) x))
                  ,instance-variables))

     ;; add user methods
     ,@(map (lambda (method)
              (let ((definer (case (car method)
                               ((define-method -) 'objc:define-method)
                               ((define-class-method +) 'objc:define-class-method)
                               (else (error "invalid method definition keyword" (car method))))))
                `(,definer ,class ,@(cdr method))))
            methods)

     ;; Add convenience methods.  The dealloc-scheme comments explain why it gets
     ;; added to every class, not just the first Scheme generation.
     (objc:add-convenience-method! ,class
                                   "dealloc"
                                   "v@:"
                                   objc_method_dealloc)))))

